/*
  my-alloc.cc
  
  Han-Wen's nifty do-it-yourself
  malloc checker.
*/


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include "my-alloc.hh"

const int MEG= 1 << 20;
const int poolsz = 20 * MEG;
const int poolidxsz = MEG;

static char *my_pool;
static char *pool_end;
struct Segment_description
{
  char *segment_ptr;
  char freed;
  // store origin?
};

static Segment_description * segments;

static int segment_array_idx;

void
init_my_allocators ()
{
  my_pool = (char*)malloc (poolsz);
  pool_end = my_pool + poolsz;
  
  segments = (Segment_description*)malloc (poolidxsz  * sizeof (Segment_description) );

  segments[0].segment_ptr = my_pool;
  segments[0].freed = 0;
  
}

void *
my_malloc(size_t s)
{
  if (!s)			// we want unique adresses. 
    s ++;
  char * p = segments[segment_array_idx].segment_ptr;
  char *ep = p + s;

  assert (ep < pool_end);

  char *scramble = p;
  while (scramble < ep)
    {
      *scramble++ = 0x55;
    }

  segments[segment_array_idx].freed = 0;
  
  segment_array_idx++;
  assert (segment_array_idx < poolidxsz);


  segments[segment_array_idx].segment_ptr = ep;
  return p;
}

int
find_segment_idx (char *p)
{
  int lo = 0;
  int hi = segment_array_idx;
  int cmp;
  /* binary search */
  do
  {
      cmp = (lo + hi) / 2;

      int result = p - segments[cmp].segment_ptr;

      if (result < 0)
          hi = cmp;
      else
          lo = cmp;
    }
  while (hi - lo > 1);
  return lo;
}

void
my_free (void *vp)
{
  char * p = (char*)vp;
  int idx = find_segment_idx (p);


  assert (segments[idx].segment_ptr == p);
  assert (!segments[idx].freed);

  segments[idx].freed = 1;
  
  char * cp = (char*)p;
  char * ep = segments[idx + 1].segment_ptr;
  
  while (cp < ep)
    {
      *cp  = 0x66;
      cp ++;
    }
}


void* operator new (size_t s)
{
  return my_malloc (s);
}

void operator delete (void *p)
{
  my_free (p);
}

void *operator new[] (size_t s)
{
  return my_malloc (s);
}

void operator delete[] (void *p)
{
  my_free (p);
}
